# coding: utf-8
u"""Паки и экшены для окна реестра "Роли"."""
from operator import or_

from django.db.models import Case
from django.db.models import F
from django.db.models import Q
from django.db.models import Value
from django.db.models import When
from django.db.models.fields import CharField
from django.db.models.functions import Concat
from django.utils.functional import cached_property
from educommon import ioc
from educommon.m3 import convert_validation_error_to
from m3.actions import ApplicationLogicException
from m3.actions import ControllerCache
from m3.actions import PreJsonResult
from m3.actions.results import OperationResult
from m3_django_compat import atomic
from m3_django_compat import get_request_params
from objectpack.actions import BaseAction
from objectpack.actions import ObjectPack
from objectpack.actions import ObjectRowsAction
from objectpack.actions import ObjectSelectWindowAction
from objectpack.exceptions import ValidationError
from objectpack.models import VirtualModel
from objectpack.tools import extract_int_list
from objectpack.tree_object_pack.actions import TreeObjectPack
from objectpack.ui import ModelEditWindow

from . import ui
from .manager import rbac
from .models import Permission
from .models import Role
from .models import RoleParent
from .models import RolePermission


def _get_role(role_id):
    try:
        return Role.objects.get(pk=role_id)
    except Role.DoesNotExist:
        raise ApplicationLogicException(u'Роль ID:{} не существует')


class RolesTreeRowsAction(ObjectRowsAction):

    u"""Экшн, отдающий данные иерархии ролей.

    Т.к. одна и та же роль может быть включена в несколько ролей, то она также
    может отображаться вложенной в несколько ролей.
    """

    @cached_property
    def _parent_ids(self):
        u"""Id ролей, содержащих в себе другие роли.

        :rtype: set
        """
        result = RoleParent.objects.values_list('parent', flat=True).distinct()

        return set(result)

    def is_leaf(self, role):
        u"""Возвращает True, если данная роль не включает другие роли.

        :param role: Роль.
        :type role: Role

        :rtype: bool
        """
        if get_request_params(self.request).get('filter', False):
            result = True
        else:
            result = role.id not in self._parent_ids

        return result

    def prepare_object(self, obj):
        u"""Сохранение данных роли в словарь перед сериализацией в JSON."""
        data = super(RolesTreeRowsAction, self).prepare_object(obj)

        data['leaf'] = self.is_leaf(obj)

        return data

    def run(self, *args, **kwargs):
        result = super(RolesTreeRowsAction, self).run(*args, **kwargs)

        data = result.data.get('rows', [])

        return PreJsonResult(data)


class AddRoleToRoleWindowAction(ObjectSelectWindowAction):

    u"""Отображение окна добавления одной роли в другую."""

    def create_window(self):
        self.win = ui.RoleSelectWindow()

    def set_window_params(self):
        super(AddRoleToRoleWindowAction, self).set_window_params()

        self.win_params['pack'] = self.parent

        role_id = getattr(self.context, self.parent.id_param_name)
        role = _get_role(role_id)
        self.win_params['role'] = role


class AddRoleToRoleAction(BaseAction):

    u"""Добавление одной роли к другой."""

    @convert_validation_error_to(ApplicationLogicException)
    def run(self, request, context):
        role_parent = RoleParent(
            role=_get_role(getattr(context, self.parent.id_param_name)),
            parent=_get_role(context.parent_id),
        )
        role_parent.full_clean()
        role_parent.save()

        return OperationResult()


class DeleteRoleFromRoleAction(BaseAction):

    u"""Удаление одной роли из другой."""

    def run(self, request, context):
        try:
            role_parent = RoleParent.objects.get(
                role=_get_role(getattr(context, self.parent.id_param_name)),
                parent=_get_role(context.parent_id),
            )
        except RoleParent.DoesNotExist:
            pass

        role_parent.delete()

        return OperationResult()


class Pack(TreeObjectPack):

    u"""Пак окна реестра "Роли"."""

    model = Role
    title = u'Роли'

    columns = [
        dict(
            data_index='name',
            header=u'Название',
            searchable=True,
        ),
    ]

    list_window = ui.RolesListWindow
    add_window = ModelEditWindow.fabricate(
        model,
        field_list=('name', 'description', 'can_be_assigned'),
        model_register=ioc.get('observer'),
    )
    edit_window = ui.RoleEditWindow

    column_name_on_select = 'name'

    def __init__(self):
        super(Pack, self).__init__()

        self.replace_action('rows_action', RolesTreeRowsAction())

        self.add_role_to_role_window_action = AddRoleToRoleWindowAction()
        self.add_role_to_role_action = AddRoleToRoleAction()
        self.delete_role_from_role_action = DeleteRoleFromRoleAction()

        self.actions.extend((
            self.add_role_to_role_window_action,
            self.add_role_to_role_action,
            self.delete_role_from_role_action,
        ))
        # ---------------------------------------------------------------------
        # Настройка разрешений для экшенов пака.
        self.need_check_permission = True
        self.perm_code = 'role'

        for action in (self.autocomplete_action,
                       self.list_window_action,
                       self.multi_select_window_action,
                       self.rows_action,
                       self.select_window_action,
                       self.edit_window_action):
            action.perm_code = 'view'

        for action in (self.new_window_action,
                       self.save_action,
                       self.delete_action,
                       self.add_role_to_role_window_action,
                       self.add_role_to_role_action,
                       self.delete_role_from_role_action):
            action.perm_code = 'edit'
        # ---------------------------------------------------------------------

    def extend_menu(self, menu):
        u"""Размещает в меню Пуск ссылку Администрирование --> Роли."""
        return menu.administry(
            menu.Item(self.title, self.list_window_action),
        )

    def declare_context(self, action):
        result = super(Pack, self).declare_context(action)

        if action in (self.rows_action,
                      self.add_role_to_role_window_action,
                      self.add_role_to_role_action):
            result[self.id_param_name] = dict(type='int')

        if action is self.rows_action:
            result['select_mode'] = dict(type='boolean', default=False)
            result['role_id'] = dict(type='int_or_none', default=None)

        if action is self.add_role_to_role_action:
            result['parent_id'] = dict(type='int')

        if action is self.save_action:
            result.update(
                name=dict(type='unicode'),
                description=dict(type='unicode', default=u''),
                permissions=dict(type='int_list', default=[]),
            )

        return result

    def get_rows_query(self, request, context):
        request_params = get_request_params(request)
        if request_params.get('filter'):
            return super(Pack, self).get_rows_query(request, context)

        current_role_id = getattr(context, self.id_param_name)

        if request_params.get('filter', False):
            result = super(Pack, self).get_rows_query(
                request, context
            )
        elif current_role_id < 0:
            # Вывод корневых ролей
            result = self.model.objects.exclude(
                pk__in=RoleParent.objects.values('role').distinct()
            )
        else:
            # Вывод ролей, вложенных в указанную роль
            result = self.model.objects.filter(
                pk__in=RoleParent.objects.filter(
                    parent=current_role_id
                ).values('role').distinct()
            )

        if context.select_mode and context.role_id is not None:
            # Режим выбора роли, в которую будет добавлена указанная роль.
            # В этом режиме надо исключить возможность создания циклов в
            # иерархии ролей, поэтому из результатов запроса надо исключить
            # все подчиненные роли.
            try:
                role = Role.objects.get(pk=context.role_id)
            except Role.DoesNotExist:
                raise ApplicationLogicException(
                    u'Роль ID:{} не найдена'.format(context.role_id)
                )

            result = result.exclude(
                pk__in=set([role.id]) | set(r.id for r in role.subroles)
            )

        return result

    def get_list_window_params(self, params, request, context):
        params = super(Pack, self).get_list_window_params(
            params, request, context
        )

        if not params['is_select_mode']:
            params['width'] = 700
            params['height'] = 600

        params['can_edit'] = rbac.has_access(self.save_action, request)

        return params

    def get_edit_window_params(self, params, request, context):
        params = super(Pack, self).get_edit_window_params(
            params, request, context
        )

        params['partitions_pack'] = ControllerCache.find_pack(PartitionsPack)
        params['permissions_pack'] = ControllerCache.find_pack(PermissionsPack)

        if not params['create_new']:
            params['permission_ids'] = list(
                params['object'].permissions.values_list('pk', flat=True)
            )

        params['can_edit'] = rbac.has_access(self.save_action, request)
        if not params['can_edit']:
            params['title'] = self.format_window_title(u'Просмотр')

        return params

    @convert_validation_error_to(ValidationError)
    @atomic
    def save_row(self, obj, create_new, request, context):
        # Сохранение роли
        obj.full_clean()
        obj.save()

        # Сохранение связи с родительской ролью
        if create_new and context.parent_id is not None:
            try:
                parent = Role.objects.get(pk=context.parent_id)
            except Role.DoesNotExist:
                raise ApplicationLogicException(
                    u'Роль ID:{} не существует.'.format(context.parent_id)
                )

            RoleParent.objects.create(role=obj, parent=parent)
        else:
            new_permissions = set(extract_int_list(request, 'permissions'))
            old_permissions = set(obj.permissions.values_list('pk', flat=True))

            # Удаление прав у роли
            for link in RolePermission.objects.filter(
                role=obj,
                permission__in=old_permissions - new_permissions,
            ):
                link.delete()

            # Добавление новых прав в роль
            for permission_id in new_permissions - old_permissions:
                RolePermission.objects.get_or_create(
                    role=obj, permission_id=permission_id
                )


class Partition(VirtualModel):

    u"""Виртуальная модель "Раздел системы".

    Используется в связи с тем, что сведения о разделах не сохраняются в БД.
    """

    def __init__(self, data):
        self.__dict__.update(data)

    @classmethod
    def _get_ids(cls):
        if not hasattr(cls, 'data'):
            cls.data = []
            for i, title in enumerate(sorted(rbac.partitions.iterkeys())):
                cls.data.append(dict(id=i, title=title))

        return cls.data

    def __unicode__(self):
        return self.title


class PartitionsPack(ObjectPack):

    u"""Пак для грида "Разделы системы" окна редактирования роли."""

    model = Partition

    columns = [
        dict(
            data_index='__unicode__',
            header=u'Модуль',
        ),
    ]

    allow_paging = False


class PermissionsPack(ObjectPack):

    u"""Пак для грида "Права доступа" окна редактирования роли."""

    model = Permission

    columns = (
        dict(
            data_index='title_with_group',
            header=u'Разрешение',
        ),
    )
    list_sort_order = ('title_with_group',)

    allow_paging = False

    def declare_context(self, action):
        result = super(PermissionsPack, self).declare_context(action)

        if action is self.rows_action:
            result['partition_id'] = dict(type='int')

        return result

    def get_rows_query(self, request, context):
        # Определение название раздела по его id.
        try:
            partition = PartitionsPack.model.objects.get(
                id=context.partition_id
            ).title
        except PartitionsPack.model.DoesNotExists:
            raise ApplicationLogicException(
                u'Раздел {} не существует'.format(context.partition_id)
            )

        query = super(PermissionsPack, self).get_rows_query(
            request, context
        ).filter(
            # Условия для выборки разрешений только из раздела partition.
            reduce(or_, (
                Q(name__startswith=code + '/')
                for code in rbac.partitions[partition]
            ))
        ).annotate(title_with_group=Case(
            # Добавление названия группы к названию разрешения.
            output_field=CharField(),
            *(
                When(
                    name__startswith=group + '/',
                    then=Concat(
                        Value(title + u' - ' if title else u''),
                        F('title'),
                    )
                )
                for group, title in rbac.groups.iteritems()
            )
        ))

        return query
